/*Copyright 2009-2011 David McCallum

This file is part of EbonHack.

    EbonHack is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EbonHack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EbonHack.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GraphicsSettings.hpp"
#include "WhiteBoard.hpp"
#include "ConfigWriter.hpp"
#include "NethackFX.hpp"
#include <QDir>

using namespace std;

GraphicsSettings::GraphicsSettings(QWidget* parent,
                                   Qt::WindowFlags flags,
                                   WhiteBoard *newWhiteBoard) : QDialog(parent, flags)
{
    QString tilesetDir;//the directory containing built-in tilesets

    whiteBoard = newWhiteBoard;

    tileChooser = new QFileDialog;
    openGLAvailable = false;
    tilesetChanged = false;
    graphicsModeChanged = false;

    gui.setupUi(this);

    //### Connect buttons ###
    connect(gui.okButton, SIGNAL(clicked()),
            this, SLOT(okClicked()));
    connect(gui.cancelButton, SIGNAL(clicked()),
            this, SLOT(cancelClicked()));
    connect(gui.defaultTilesetButton, SIGNAL(toggled(bool)),
            this, SLOT(defaultToggled(bool)));
    connect(gui.customTilesetButton, SIGNAL(toggled(bool)),
            this, SLOT(customToggled(bool)));
    connect(gui.selectTilesetButton, SIGNAL(clicked()),
            this, SLOT(selectTilesetClicked()));
    connect(gui.nativeQtButton, SIGNAL(toggled(bool)),
            this, SLOT(nativeQtToggled(bool)));
    connect(gui.openGLButton, SIGNAL(toggled(bool)),
            this, SLOT(openGLToggled(bool)));

    //### Configure the tile chooser ###
    connect(tileChooser, SIGNAL(fileSelected(const QString&)),
            this, SLOT(tilesetChosen(const QString&)));
    tileChooser->setFileMode(QFileDialog::ExistingFile);

    tilesetDir = QDir::current().absolutePath() + "/data/tilesets";
    tileChooser->setDirectory(tilesetDir);

    tileChooser->setNameFilter("*.bmp *.png *.jpg *.jpeg *.gif *.tiff");

    //### Load graphics configuration ###
    load();
}//constructor

GraphicsSettings::~GraphicsSettings(void)
{
    delete tileChooser;
    tileChooser = NULL;
}//destructor

void GraphicsSettings::load(void)
{
    string tileset = ConfigWriter::loadString("game_config.txt", "Tileset");
    bool useTileset = ConfigWriter::loadBool("game_config.txt", "Use Tileset");
    bool useOpenGL = ConfigWriter::loadBool("game_config.txt", "Use OpenGL");
    bool useServerTiles = ConfigWriter::loadBool("game_config.txt", "Use Server Tiles");

    //### Disable the OpenGL button if not available ###
    if (openGLAvailable)
        gui.openGLButton->setEnabled(true);
    else
        gui.openGLButton->setEnabled(false);

    //### Check the proper graphics button ###
    if ((useOpenGL) && (openGLAvailable))
        gui.openGLButton->setChecked(true);
    else
        gui.nativeQtButton->setChecked(true);

    //### Check the custom tileset button ###
    if (useTileset)
    {
        gui.customTilesetButton->setChecked(true);
        gui.tilesetBox->setEnabled(true);
        gui.selectTilesetButton->setEnabled(true);
    }//if useTileset
    else
    {
        gui.customTilesetButton->setChecked(false);
        gui.tilesetBox->setEnabled(false);
        gui.selectTilesetButton->setEnabled(false);
    }//else useTileset

    //### Set the custom tileset text ###
    gui.tilesetBox->setText(QString::fromStdString(tileset));

    //### Check the Use Server Tiles button ###
    gui.serverTilesBox->setChecked(useServerTiles);

    tilesetChanged = false;
    graphicsModeChanged = false;
}//load

void GraphicsSettings::setOpenGL(bool available)
{
    bool useOpenGL = ConfigWriter::loadBool("game_config.txt", "Use OpenGL");

    openGLAvailable = available;

    //### Disable OpenGL in the config file if not available ###
    if ((!available) && (useOpenGL))
    {
        useOpenGL = false;
        ConfigWriter::writeBool("game_config.txt", "Use OpenGL", false);
    }//if !available && useOpenGL

    //### Reload settings ###
    load();
}//setOpenGL

void GraphicsSettings::okClicked(void)
{
    NethackFX *netFX = whiteBoard->getNetFX();
    QString tileset = gui.tilesetBox->text();
    bool useTileset = gui.customTilesetButton->isChecked();
    bool useOpenGL = gui.openGLButton->isChecked();
    bool useServerTiles = gui.serverTilesBox->isChecked();
    bool allOK = true;//false if there's an error and we shouldn't close the dialog

    //### Make sure we have an entry to write to file ###
    if (tileset.size() == 0)
    {
        tileset = "NONE";
        gui.tilesetBox->setText(tileset);
    }//if size()

    //### Load the new tileset ###
    if (tilesetChanged)
    {
        if (useTileset)
        {
            if (!netFX->loadCustomTiles(tileset))
                allOK = false;
        }//if useTileset
        else
        {
            if (!netFX->loadDefaultTiles())
                allOK = false;
        }//else useTileset
    }//if tilesetChanged

    //### Save settings and close the dialog ###
    if (allOK)
    {
        if (graphicsModeChanged)
        {
            whiteBoard->showMessage("EbonHack must be restarted for the changes to take effect.");
            graphicsModeChanged = false;
        }//if graphicsModeChanged

        tilesetChanged = false;

        ConfigWriter::writeBool("game_config.txt", "Use Tileset", useTileset);
        ConfigWriter::writeString("game_config.txt", "Tileset", tileset.toStdString());
        ConfigWriter::writeBool("game_config.txt", "Use OpenGL", useOpenGL);
        ConfigWriter::writeBool("game_config.txt", "Use Server Tiles", useServerTiles);

        hide();
    }//if allOK
}//okClicked

void GraphicsSettings::cancelClicked(void)
{
    load();
    hide();
}//cancelClicked

void GraphicsSettings::defaultToggled(bool checked)
{
    if (checked)
    {
        gui.tilesetBox->setEnabled(false);
        gui.selectTilesetButton->setEnabled(false);

        tilesetChanged = true;
    }//if checked
}//defaultToggled

void GraphicsSettings::customToggled(bool checked)
{
    if (checked)
    {
        gui.tilesetBox->setEnabled(true);
        gui.selectTilesetButton->setEnabled(true);

        tilesetChanged = true;
    }//if checked
}//customToggled

void GraphicsSettings::nativeQtToggled(bool)
{
    graphicsModeChanged = true;
}//nativeQtToggled

void GraphicsSettings::openGLToggled(bool)
{
    graphicsModeChanged = true;
}//openGLToggled

void GraphicsSettings::selectTilesetClicked(void)
{
    tileChooser->show();
}//selectTilesetClicked

void GraphicsSettings::tilesetChosen(const QString &filename)
{
    tilesetChanged = true;
    gui.tilesetBox->setText(filename);
}//tilesetChosen

bool GraphicsSettings::openGLEnabled(void)
{
    return gui.openGLButton->isChecked();
}//openGLEnabled

bool GraphicsSettings::serverTilesEnabled(void)
{
    return gui.serverTilesBox->isChecked();
}//serverTilesEnabled
